

// Adapted from the haste library
//
// See:
// Copyright 2020 LMNT, Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// ==============================================================================

#pragma once

#include "../util/util.h"

#ifndef FLASHRNN_NUM_GATES_R
// all needed definitions from external
#define FLASHRNN_NUM_HEADS 1
#define FLASHRNN_HIDDEN_SIZE 512
#define FLASHRNN_BATCH_SIZE 8
#define FLASHRNN_NUM_GATES_R 4
#define FLASHRNN_NUM_GATES_W 4
#define FLASHRNN_NUM_GATES_I 4
#define FLASHRNN_NUM_GATES_T 4
#define FLASHRNN_NUM_STATES 4
#define FLASHRNN_DTYPE_R float
#define FLASHRNN_DTYPE_B float
#define FLASHRNN_DTYPE_W float
#define FLASHRNN_DTYPE_G float
#define FLASHRNN_DTYPE_S float
#define FLASHRNN_DTYPE_A float
// defines whether g = Wx + Ry + b for every gate, enables half the cache for
// backward
#define FLASHRNN_SIMPLE_AGG true
#endif

namespace flashrnn {

class ForwardPass {
public:
  // training: `true` if the caller intends to perform a backward pass to
  // compute gradients
  ForwardPass(const bool training, const int batch_size, const int hidden_size,
              const int num_heads, const cublasHandle_t &blas_handle,
              const cudaStream_t &stream = 0);

  // Releases internal resources.
  // Blocks until all iterations have completed executing on the GPU.
  ~ForwardPass();

  // Set internal values for single forward / backward
  void Set(const bool training, const int batch_size, const int hidden_size,
           const int num_heads, const cublasHandle_t &blas_handle,
           const cudaStream_t &stream = 0);

  // Performs one forward time step iteration of the RNN.
  int Iterate(const cudaStream_t &stream, const FLASHRNN_DTYPE_R *R,
              const FLASHRNN_DTYPE_B *b, const FLASHRNN_DTYPE_W *x,
              const FLASHRNN_DTYPE_S *s, FLASHRNN_DTYPE_S *s_out,
              FLASHRNN_DTYPE_G *g_r, FLASHRNN_DTYPE_G *g_i,
              FLASHRNN_DTYPE_G *tmp_Ry);

  int Run(const int steps, const FLASHRNN_DTYPE_R *R, const FLASHRNN_DTYPE_B *b,
          const FLASHRNN_DTYPE_W *x, FLASHRNN_DTYPE_S *s, FLASHRNN_DTYPE_G *g_r,
          FLASHRNN_DTYPE_G *g_i, FLASHRNN_DTYPE_G *tmp_Ry);

private:
  int IterateInternal(const FLASHRNN_DTYPE_W *x, const FLASHRNN_DTYPE_R *R,
                      const FLASHRNN_DTYPE_B *b, const FLASHRNN_DTYPE_S *s,
                      const uint s_stride, FLASHRNN_DTYPE_S *s_out,
                      const uint s_out_stride, FLASHRNN_DTYPE_G *g_r,
                      FLASHRNN_DTYPE_G *g_i, FLASHRNN_DTYPE_G *tmp_Ry);

  struct private_data;
  private_data *data_;
};

class BackwardPass {
public:
  BackwardPass(const int batch_size, const int hidden_size, const int num_heads,
               const cublasHandle_t &blas_handle,
               const cudaStream_t &stream = 0);

  // Set internal values for single forward / backward
  void Set(const int batch_size, const int hidden_size, const int num_heads,
           const cublasHandle_t &blas_handle, const cudaStream_t &stream = 0);

  // Releases internal resources.
  // Blocks until all iterations have completed executing on the GPU.
  ~BackwardPass();

  int Iterate(const cudaStream_t &stream, const FLASHRNN_DTYPE_R *R_t,
              const FLASHRNN_DTYPE_B *b, const FLASHRNN_DTYPE_S *s,
              const FLASHRNN_DTYPE_S *s_new, const FLASHRNN_DTYPE_S *ds_new,
              FLASHRNN_DTYPE_R *dR, FLASHRNN_DTYPE_B *db, FLASHRNN_DTYPE_S *ds,
              FLASHRNN_DTYPE_G *g_r, FLASHRNN_DTYPE_G *g_i,
              FLASHRNN_DTYPE_G *g_b);

  int Run(const int steps, const FLASHRNN_DTYPE_R *R_t,
          const FLASHRNN_DTYPE_B *b, const FLASHRNN_DTYPE_S *s,
          const FLASHRNN_DTYPE_S *ds_new, FLASHRNN_DTYPE_R *dR,
          FLASHRNN_DTYPE_B *db, FLASHRNN_DTYPE_S *ds, FLASHRNN_DTYPE_G *g_r,
          FLASHRNN_DTYPE_G *g_i, FLASHRNN_DTYPE_G *g_b);

private:
  int IterateInternal(const FLASHRNN_DTYPE_R *R_t, const FLASHRNN_DTYPE_B *b,
                      const FLASHRNN_DTYPE_S *s, const uint s_stride,
                      const FLASHRNN_DTYPE_S *s_new, const uint s_new_stride,
                      const FLASHRNN_DTYPE_S *ds_new, const uint ds_new_stride,
                      FLASHRNN_DTYPE_S *ds, const uint ds_stride,
                      FLASHRNN_DTYPE_G *g_r, FLASHRNN_DTYPE_G *g_i,
                      FLASHRNN_DTYPE_G *g_b);
  struct private_data;
  private_data *data_;
};

class BackwardPassCut {
public:
  BackwardPassCut(const int batch_size, const int hidden_size,
                  const int num_heads, const cublasHandle_t &blas_handle,
                  const cudaStream_t &stream = 0);

  // Set internal values for single forward / backward
  void Set(const int batch_size, const int hidden_size, const int num_heads,
           const cublasHandle_t &blas_handle, const cudaStream_t &stream = 0);

  // Releases internal resources.
  // Blocks until all iterations have completed executing on the GPU.
  ~BackwardPassCut();

  int Iterate(const cudaStream_t &stream, const FLASHRNN_DTYPE_R *R_t,
              const FLASHRNN_DTYPE_B *b, const FLASHRNN_DTYPE_S *s,
              const FLASHRNN_DTYPE_S *s_new, const FLASHRNN_DTYPE_S *ds_new,
              FLASHRNN_DTYPE_R *dR, FLASHRNN_DTYPE_B *db, FLASHRNN_DTYPE_S *ds,
              FLASHRNN_DTYPE_G *g_r, FLASHRNN_DTYPE_G *g_i,
              FLASHRNN_DTYPE_G *g_bias);

  int Run(const int steps, const FLASHRNN_DTYPE_R *R_t,
          const FLASHRNN_DTYPE_B *b, const FLASHRNN_DTYPE_S *s,
          const FLASHRNN_DTYPE_S *ds_new, FLASHRNN_DTYPE_R *dR,
          FLASHRNN_DTYPE_B *db, FLASHRNN_DTYPE_S *ds, FLASHRNN_DTYPE_G *g_r,
          FLASHRNN_DTYPE_G *g_i, FLASHRNN_DTYPE_G *g_bias);

private:
  int IterateInternal(const FLASHRNN_DTYPE_R *R_t, const FLASHRNN_DTYPE_B *b,
                      const FLASHRNN_DTYPE_S *s, const uint s_stride,
                      const FLASHRNN_DTYPE_S *s_new, const uint s_new_stride,
                      const FLASHRNN_DTYPE_S *ds_new, const uint ds_new_stride,
                      FLASHRNN_DTYPE_S *ds, const uint ds_stride,
                      FLASHRNN_DTYPE_G *g_r, FLASHRNN_DTYPE_G *g_i,
                      FLASHRNN_DTYPE_G *g_bias);
  struct private_data;
  private_data *data_;
};

} // namespace flashrnn
